#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

def np():
    import numpy
    return numpy

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.signature import Lifespan
from titus.signature import PFAVersion
from titus.datatype import *
from titus.errors import *
from titus.util import callfcn, div, flatten
import titus.P as P
from titus.lib.array import argLowestN
from titus.lib.prob.dist import Chi2Distribution

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "model.reg."

#################################################################### 

class Linear(LibFcn):
    name = prefix + "linear"
    sig  = Sigs([Sig([{"datum": P.Array(P.Double())}, {"model": P.WildRecord("M", {"coeff": P.Array(P.Double()), "const": P.Double()})}], P.Double()),
                 Sig([{"datum": P.Array(P.Double())}, {"model": P.WildRecord("M", {"coeff": P.Array(P.Array(P.Double())), "const": P.Array(P.Double())})}], P.Array(P.Double())),
                 Sig([{"datum": P.Map(P.Double())}, {"model": P.WildRecord("M", {"coeff": P.Map(P.Double()), "const": P.Double()})}], P.Double()),
                 Sig([{"datum": P.Map(P.Double())}, {"model": P.WildRecord("M", {"coeff": P.Map(P.Map(P.Double())), "const": P.Map(P.Double())})}], P.Map(P.Double()))])
    errcodeBase = 31000
    def __call__(self, state, scope, pos, paramTypes, datum, model):
        coeffType = [x["type"] for x in paramTypes[1]["fields"] if x["name"] == "coeff"][0]

        if coeffType == {'items': 'double', 'type': 'array'}: #sig1
            coeff = model["coeff"] + [model["const"]]
            datum = np().array(datum + [1.0])
            if len(datum) != len(coeff):
                raise PFARuntimeException("misaligned coeff", self.errcodeBase + 0, self.name, pos)
            return float(np().dot(coeff, datum))

        elif coeffType == {'items': {'items': 'double', 'type': 'array'}, 'type': 'array'}: #sig2
            if len(model["const"]) != len(model["coeff"]):
                raise PFARuntimeException("misaligned const", self.errcodeBase + 1, self.name, pos)
            elif any(len(row) != len(datum) for row in model["coeff"]):
                raise PFARuntimeException("misaligned coeff", self.errcodeBase + 0, self.name, pos)
            elif len(datum) == 0:
                return model["const"]
            elif len(model["const"]) == 0:
                return []
            else:
                coeff = np().array(model["coeff"])
                const = np().array(model["const"])
                datum = np().array(datum + [1.0])
                coeff = np().vstack((coeff.T, const))
                return list(map(float, np().dot(coeff.T, datum)))

        elif coeffType == {'values': 'double', 'type': 'map'}: #sig3
            coeff = model["coeff"]
            const = model["const"]
            out = 0.0
            for key in set(list(datum.keys()) + list(coeff.keys())):
                out += datum.get(key, 0.0) * coeff.get(key, 0.0)
            return float(out + const)

        else:
            coeff = model["coeff"]
            const = model["const"]
            outMap = {}

            innerKeys = set(list(datum.keys()) + sum([list(x.keys()) for x in list(coeff.values())], []))
            outerKeys = set(list(const.keys()) + list(coeff.keys()))

            for outerKey in outerKeys:
                out = 0.0
                for innerKey in innerKeys:
                    out += datum.get(innerKey, 0.0) * coeff.get(outerKey, {}).get(innerKey, 0.0)
                outMap[outerKey] = float(out + const.get(outerKey, 0.0))
            return outMap

provide(Linear())

class LinearVariance(LibFcn):
    name = prefix + "linearVariance"
    sig = Sigs([Sig([{"datum": P.Array(P.Double())}, {"model": P.WildRecord("M", {"covar": P.Array(P.Array(P.Double()))})}], P.Double()),
                Sig([{"datum": P.Array(P.Double())}, {"model": P.WildRecord("M", {"covar": P.Array(P.Array(P.Array(P.Double())))})}], P.Array(P.Double())),
                Sig([{"datum": P.Map(P.Double())}, {"model": P.WildRecord("M", {"covar": P.Map(P.Map(P.Double()))})}], P.Double()),
                Sig([{"datum": P.Map(P.Double())}, {"model": P.WildRecord("M", {"covar": P.Map(P.Map(P.Map(P.Double())))})}], P.Map(P.Double()))])
    errcodeBase = 31010
    def __call__(self, state, scope, pos, paramTypes, datum, model):
        covarType = [x["type"] for x in paramTypes[1]["fields"] if x["name"] == "covar"][0]

        if covarType == {"type": "array", "items": {"type": "array", "items": "double"}}:  # sig1
            datum = datum + [1.0]
            covar = model["covar"]
            if len(datum) != len(covar) or any(len(datum) != len(x) for x in covar):
                raise PFARuntimeException("misaligned covariance", self.errcodeBase + 0, self.name, pos)
            x = np().matrix([datum])
            C = np().matrix(covar)
            return float(x.dot(C.dot(x.T))[0][0])

        elif covarType == {"type": "array", "items": {"type": "array", "items": {"type": "array", "items": "double"}}}:  # sig2
            datum = datum + [1.0]
            x = np().matrix([datum])
            out = []
            for covar in model["covar"]:
                if len(datum) != len(covar) or any(len(datum) != len(x) for x in covar):
                    raise PFARuntimeException("misaligned covariance", self.errcodeBase + 0, self.name, pos)
                C = np().matrix(covar)
                out.append(float(x.dot(C.dot(x.T))[0][0]))
            return out

        elif covarType == {"type": "map", "values": {"type": "map", "values": "double"}}:  # sig3
            datum = dict(list(datum.items()) + [("", 1.0)])
            covar = model["covar"]
            keys = list(set(list(datum.keys()) + sum([list(x.keys()) for x in list(covar.values())], [])))
            x = np().matrix([[datum.get(k, 0.0) for k in keys]])
            C = np().matrix([[covar.get(i, {}).get(j, 0.0) for j in keys] for i in keys])
            return float(x.dot(C.dot(x.T))[0][0])

        else:
            datum = dict(list(datum.items()) + [("", 1.0)])
            covar = model["covar"]
            keys = list(set(list(datum.keys()) + sum(flatten([[list(x.keys()) for x in list(row.values())] for row in list(covar.values())]), [])))
            x = np().matrix([[datum.get(k, 0.0) for k in keys]])
            out = {}
            for depkey, row in list(covar.items()):
                C = np().matrix([[row.get(i, {}).get(j, 0.0) for j in keys] for i in keys])
                out[depkey] = float(x.dot(C.dot(x.T))[0][0])
            return out

provide(LinearVariance())

class GaussianProcess(LibFcn):
    name = prefix + "gaussianProcess"
    sig = Sigs([Sig([{"x": P.Double()}, {"table": P.Array(P.WildRecord("R", {"x": P.Double(), "to": P.Double()}))},
                     {"krigingWeight": P.Union([P.Null(), P.Double()])},
                     {"kernel": P.Fcn([P.Array(P.Double()), P.Array(P.Double())], P.Double())}], P.Double()),

                Sig([{"x": P.Double()}, {"table": P.Array(P.WildRecord("R", {"x": P.Double(), "to": P.Array(P.Double())}))},
                     {"krigingWeight": P.Union([P.Null(), P.Double()])},
                     {"kernel": P.Fcn([P.Array(P.Double()), P.Array(P.Double())], P.Double())}], P.Array(P.Double())),

                Sig([{"x": P.Array(P.Double())}, {"table": P.Array(P.WildRecord("R", {"x": P.Array(P.Double()), "to": P.Double()}))},
                     {"krigingWeight": P.Union([P.Null(), P.Double()])},
                     {"kernel": P.Fcn([P.Array(P.Double()), P.Array(P.Double())], P.Double())}], P.Double()),

                Sig([{"x": P.Array(P.Double())}, {"table": P.Array(P.WildRecord("R", {"x": P.Array(P.Double()), "to": P.Array(P.Double())}))},
                     {"krigingWeight": P.Union([P.Null(), P.Double()])},
                     {"kernel": P.Fcn([P.Array(P.Double()), P.Array(P.Double())], P.Double())}], P.Array(P.Double()))])

    errcodeBase = 31080

    def genpy(self, paramTypes, args, pos):
        toType = None
        for x in paramTypes[1].items.fields:
            if x.name == "to":
                toType = x.avroType
        for x in paramTypes[1].items.fields:
            if x.name == "sigma":
                if not toType.accepts(x.avroType) or not x.avroType.accepts(toType):
                    raise PFASemanticException(self.name + " is being given a table record in which the \"sigma\" field does not have the same type as the \"to\" field: " + str(x.avroType), None)
        return LibFcn.genpy(self, paramTypes, args, pos)

    def getbeta(self, krigingWeight, pos):
        if isinstance(krigingWeight, dict):
            beta = krigingWeight["double"]
            if math.isnan(beta) or math.isinf(beta):
                raise PFARuntimeException("krigingWeight is not finite", self.errcodeBase + 7, self.name, pos)
        else:
            beta = krigingWeight

    def getnoutputs(self, x, table, paramTypes, pos):
        if isinstance(paramTypes[-1], dict) and paramTypes[-1].get("type") == "array":
            n_outputs = len(table[0]["to"])
            if n_outputs < 1:
                raise PFARuntimeException("table outputs must have at least 1 dimension", self.errcodeBase + 3, self.name, pos)

            if any(len(t["to"]) != n_outputs for t in table):
                raise PFARuntimeException("table outputs must all have the same number of dimensions", self.errcodeBase + 4, self.name, pos)

            if isinstance(x, (list, tuple)):
                if any(math.isnan(xi) or math.isinf(xi) for xi in x):
                    raise PFARuntimeException("x is not finite", self.errcodeBase + 5, self.name, pos)
            else:
                if math.isnan(x) or math.isinf(x):
                    raise PFARuntimeException("x is not finite", self.errcodeBase + 5, self.name, pos)

            if any(any(math.isnan(ti) or math.isinf(ti) for ti in t["to"]) for t in table):
                raise PFARuntimeException("table value is not finite", self.errcodeBase + 6, self.name, pos)

        else:
            if isinstance(x, (list, tuple)):
                if any(math.isnan(xi) or math.isinf(xi) for xi in x):
                    raise PFARuntimeException("x is not finite", self.errcodeBase + 5, self.name, pos)
            else:
                if math.isnan(x) or math.isinf(x):
                    raise PFARuntimeException("x is not finite", self.errcodeBase + 5, self.name, pos)

            if any(math.isnan(t["to"]) or math.isinf(t["to"]) for t in table):
                raise PFARuntimeException("table value is not finite", self.errcodeBase + 6, self.name, pos)
            n_outputs = None

        return n_outputs

    def __call__(self, state, scope, pos, paramTypes, x, table, krigingWeight, kernel):
        def kern(xvector, yvector):
            return callfcn(state, scope, kernel, [xvector, yvector])

        n_samples = len(table)
        if n_samples < 1:
            raise PFARuntimeException("table must have at least 1 entry", self.errcodeBase + 0, self.name, pos)
                
        if isinstance(x, (list, tuple)):
            n_features = len(x)
            if n_features < 1:
                raise PFARuntimeException("x must have at least 1 feature", self.errcodeBase + 1, self.name, pos)
            if any(len(t["x"]) != n_features for t in table):
                raise PFARuntimeException("table must have the same number of features as x", self.errcodeBase + 2, self.name, pos)

            n_outputs = self.getnoutputs(x, table, paramTypes, pos)

            if any(any(math.isnan(xi) or math.isinf(xi) for xi in t["x"]) for t in table):
                raise PFARuntimeException("table value is not finite", self.errcodeBase + 6, self.name, pos)

            beta = self.getbeta(krigingWeight, pos)

            X = np().array([t["x"] for t in table])
            
            if n_outputs is None:
                y = np().array([t["to"] for t in table])
                if "sigma" in table[0]:
                    nugget = np().array([(t["sigma"]/t["to"])**2 if t["to"] != 0.0 else float("inf") for t in table])
                else:
                    nugget = 10.0 * np().finfo(np().double).eps

                beta, gamma = self.fit(X, y, beta, nugget, kern, pos)
                return self.predict(x, X, y, beta, gamma, kern)

            else:
                out = [None] * n_outputs
                for i in range(n_outputs):
                    y = np().array([t["to"][i] for t in table])
                    if "sigma" in table[0]:
                        nugget = np().array([(t["sigma"][i]/t["to"][i])**2 if t["to"][i] != 0.0 else float("inf") for t in table])
                    else:
                        nugget = 10.0 * np().finfo(np().double).eps

                    beta, gamma = self.fit(X, y, beta, nugget, kern, pos)
                    out[i] = self.predict(x, X, y, beta, gamma, kern)

                return out

        else:
            n_outputs = self.getnoutputs(x, table, paramTypes, pos)

            if any(math.isnan(t["x"]) or math.isinf(t["x"]) for t in table):
                raise PFARuntimeException("table value is not finite", self.errcodeBase + 6, self.name, pos)

            beta = self.getbeta(krigingWeight, pos)

            X = np().array([[t["x"]] for t in table])

            if n_outputs is None:
                y = np().array([t["to"] for t in table])
                if "sigma" in table[0]:
                    nugget = np().array([(t["sigma"]/t["to"])**2 if t["to"] != 0.0 else float("inf") for t in table])
                else:
                    nugget = 10.0 * np().finfo(np().double).eps

                beta, gamma = self.fit(X, y, beta, nugget, kern, pos)
                return self.predict(x, X, y, beta, gamma, kern)

            else:
                out = [None] * n_outputs
                for i in range(n_outputs):
                    y = np().array([t["to"][i] for t in table])
                    if "sigma" in table[0]:
                        nugget = np().array([(t["sigma"][i]/t["to"][i])**2 if t["to"][i] != 0.0 else float("inf") for t in table])
                    else:
                        nugget = 10.0 * np().finfo(np().double).eps

                    beta, gamma = self.fit(X, y, beta, nugget, kern, pos)
                    out[i] = self.predict(x, X, y, beta, gamma, kern)
                return out

    def fit(self, X, y, beta, nugget, kern, pos):
        n_samples, n_features = X.shape

        Xnorm = (X - X.mean(axis=0))/X.std(axis=0)
        ynorm = (y - y.mean(axis=0))/y.std(axis=0)

        n_nonzero_cross_dist = n_samples * (n_samples - 1) // 2
        ij = np().zeros((n_nonzero_cross_dist, 2), dtype=np().int)
        r = np().zeros(n_nonzero_cross_dist)
        ll_1 = 0
        for k in range(n_samples - 1):
            ll_0 = ll_1
            ll_1 = ll_0 + n_samples - k - 1
            ij[ll_0:ll_1, 0] = k
            ij[ll_0:ll_1, 1] = np().arange(k + 1, n_samples)
            r[ll_0:ll_1] = [kern(Xnorm[k].tolist(), XnormOther.tolist()) for XnormOther in Xnorm[(k + 1):n_samples]]

        ij = ij.astype(np().int)

        R = np().eye(n_samples) * (1. + nugget)
        R[ij[:, 0], ij[:, 1]] = r
        R[ij[:, 1], ij[:, 0]] = r

        try:
            C = np().linalg.cholesky(R)
        except numpy.linalg.linalg.LinAlgError:
            raise PFARuntimeException("matrix of kernel results is not positive definite", self.errcodeBase + 8, self.name, pos)

        F = np().array([[1.0]] * len(X))
        Ft = np().linalg.solve(C, F)

        Q, G = np().linalg.qr(Ft)
        Yt = np().linalg.solve(C, ynorm)

        # beta can be computed from universal Kriging or a given
        if beta is None:
            beta = np().linalg.solve(G, np().dot(Q.T, Yt))
        else:
            beta = np().array([beta])

        rho = Yt - np().dot(Ft, beta)

        gamma = np().linalg.solve(C.T, rho)

        return float(beta[0]), gamma.ravel()

    def predict(self, x_pred, X, y, beta, gamma, kern):
        Xmean = X.mean(axis=0)
        Xstd = X.std(axis=0)
        Xnorm = (X - Xmean) / Xstd

        x_pred_norm = (np().array(x_pred) - Xmean) / Xstd

        y_norm = beta
        for sampleX, g in zip(Xnorm, gamma):
            r = kern(x_pred_norm.tolist(), sampleX.tolist())
            y_norm += r * g

        return y.mean() + y.std() * y_norm

provide(GaussianProcess())







############################################################# deprecated stuff (moved to stat.test.*)

class Residual(LibFcn):
    name = prefix + "residual"
    sig = Sigs([Sig([{"observation": P.Double()}, {"prediction": P.Double()}], P.Double(), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.residual instead")),
                Sig([{"observation": P.Array(P.Double())}, {"prediction": P.Array(P.Double())}], P.Array(P.Double()), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.residual instead")),
                Sig([{"observation": P.Map(P.Double())}, {"prediction": P.Map(P.Double())}], P.Map(P.Double()), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.residual instead"))])
    errcodeBase = 31020
    def __call__(self, state, scope, pos, paramTypes, observation, prediction):
        if isinstance(observation, dict):
            if len(observation) != len(prediction):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            result = {}
            for k, o in list(observation.items()):
                try:
                    result[k] = o - prediction[k]
                except KeyError:
                    raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            return result

        elif isinstance(observation, (tuple, list)):
            try:
                result = [float(o - p) for o, p in zip(observation, prediction)]
            except:
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            return result

        else:
            return float(observation - prediction)

provide(Residual())

class Pull(LibFcn):
    name = prefix + "pull"
    sig = Sigs([Sig([{"observation": P.Double()}, {"prediction": P.Double()}, {"uncertainty": P.Double()}], P.Double(), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.pull instead")),
                Sig([{"observation": P.Array(P.Double())}, {"prediction": P.Array(P.Double())}, {"uncertainty": P.Array(P.Double())}], P.Array(P.Double()), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.pull instead")),
                Sig([{"observation": P.Map(P.Double())}, {"prediction": P.Map(P.Double())}, {"uncertainty": P.Map(P.Double())}], P.Map(P.Double()), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.pull instead"))])
    errcodeBase = 31030
    def __call__(self, state, scope, pos, paramTypes, observation, prediction, uncertainty):
        if isinstance(observation, dict):
            if len(observation) != len(prediction):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            if len(observation) != len(uncertainty):
                raise PFARuntimeException("misaligned uncertainty", self.errcodeBase + 1, self.name, pos)
            result = {}
            for k, o in list(observation.items()):
                try:
                    p = prediction[k]
                except KeyError:
                    raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
                try:
                    u = uncertainty[k]
                except KeyError:
                    raise PFARuntimeException("misaligned uncertainty", self.errcodeBase + 1, self.name, pos)
                try:
                    result[k] = (o - p) / u
                except ZeroDivisionError:
                    result[k] = float("nan")
            return result

        elif isinstance(observation, (tuple, list)):
            if len(observation) != len(prediction):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            if len(observation) != len(uncertainty):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 0, self.name, pos)
            return [float((o - p)/u) for o, p, u in zip(observation, prediction, uncertainty)]

        else:
            return float((observation - prediction)/uncertainty)

provide(Pull())

class Mahalanobis(LibFcn):
    name = prefix + "mahalanobis"
    sig = Sigs([Sig([{"observation": P.Array(P.Double())}, {"prediction": P.Array(P.Double())}, {"covariance": P.Array(P.Array(P.Double()))}], P.Double(), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.mahalanobis instead")),
                Sig([{"observation": P.Map(P.Double())}, {"prediction": P.Map(P.Double())}, {"covariance": P.Map(P.Map(P.Double()))}], P.Double(), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.mahalanobis instead"))])
    errcodeBase = 31040
    def __call__(self, state, scope, pos, paramTypes, observation, prediction, covariance):
        if isinstance(observation, (tuple, list)):
            if (len(observation) < 1):
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if (len(observation) != len(prediction)):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 1, self.name, pos)
            if (not all(len(i)==len(covariance[0]) for i in covariance)) and (len(covariance) != len(covariance[0])):
                raise PFARuntimeException("misaligned covariance", self.errcodeBase + 2, self.name, pos)
            x = np().array([(o - p) for o, p in zip(observation, prediction)])
            C = np().array(covariance)
        else:
            if (len(observation) < 1):
                raise PFARuntimeException("too few rows/cols", self.errcodeBase + 0, self.name, pos)
            if (len(observation) != len(prediction)):
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 1, self.name, pos)
            # use observation keys throughout
            keys = list(observation.keys())
            try:
                x = np().array([observation[key] - prediction[key] for key in keys])
            except:
                raise PFARuntimeException("misaligned prediction", self.errcodeBase + 1, self.name, pos)
            C = np().empty((len(keys), len(keys)))
            try:
                for i,k1 in enumerate(keys):
                    for j,k2 in enumerate(keys):
                        C[i,j] = float(covariance[k1][k2])
            except:
                raise PFARuntimeException("misaligned covariance", self.errcodeBase + 2, self.name, pos)
        return float(np().sqrt(np().linalg.solve(C, x).T.dot(x)))
provide(Mahalanobis())


class UpdateChi2(LibFcn):
    name = prefix + "updateChi2"
    sig = Sigs([Sig([{"pull": P.Double()}, {"state_": P.WildRecord("A", {"chi2": P.Double(), "DOF": P.Int()})}], P.Wildcard("A"), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.updateChi2 instead")),
                Sig([{"pull": P.Array(P.Double())}, {"state_": P.WildRecord("A", {"chi2": P.Double(), "DOF": P.Int()})}], P.Wildcard("A"), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.updateChi2 instead")),
                Sig([{"pull": P.Map(P.Double())}, {"state_": P.WildRecord("A", {"chi2": P.Double(), "DOF": P.Int()})}], P.Wildcard("A"), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.updateChi2 instead"))])
    errcodeBase = 31050
    def __call__(self, state, scope, pos, paramTypes, pull, state_):
        if isinstance(pull, float):
            return update(pull*pull, state_)
        elif isinstance(pull, (tuple, list)):
            return update(sum([y**2 for y in pull]), state_)
        else:
            return update(sum([y**2 for y in list(pull.values())]), state_)
provide(UpdateChi2())
def update(x, state_):
    state_["chi2"] = float(state_["chi2"] + x)
    state_["DOF"]  = float(state_["DOF"] + 1)
    return state_

class ReducedChi2(LibFcn):
    name = prefix + "reducedChi2"
    sig = Sig([{"state_": P.WildRecord("A", {"chi2": P.Double(), "DOF": P.Int()})}], P.Double(), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.reducedChi2 instead"))
    errcodeBase = 31060
    def __call__(self, state, scope, pos, paramTypes, state_):
        return float(state_["chi2"]/state_["DOF"])
provide(ReducedChi2())

class Chi2Prob(LibFcn):
    name = prefix + "chi2Prob"
    sig = Sig([{"state_": P.WildRecord("A", {"chi2": P.Double(), "DOF": P.Int()})}], P.Double(), Lifespan(None, PFAVersion(0, 7, 2), PFAVersion(0, 9, 0), "use test.chi2Prob instead"))
    errcodeBase = 31070
    def __call__(self, state, scope, pos, paramTypes, state_):
        return float(Chi2Distribution(state_["DOF"], self.name, self.errcodeBase, pos).CDF(state_["chi2"]))
provide(Chi2Prob())





#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from titus.datatype import *

class TestDataType(unittest.TestCase):
    def testPromoteNumbers(self):
        self.assertTrue(AvroInt().accepts(AvroInt()))

        self.assertTrue(AvroInt().accepts(AvroInt()))
        self.assertTrue(AvroLong().accepts(AvroInt()))
        self.assertTrue(AvroFloat().accepts(AvroInt()))
        self.assertTrue(AvroDouble().accepts(AvroInt()))

        self.assertTrue(AvroLong().accepts(AvroLong()))
        self.assertTrue(AvroFloat().accepts(AvroLong()))
        self.assertTrue(AvroDouble().accepts(AvroLong()))

        self.assertTrue(AvroFloat().accepts(AvroFloat()))
        self.assertTrue(AvroDouble().accepts(AvroFloat()))

        self.assertTrue(AvroDouble().accepts(AvroDouble()))

        self.assertFalse(AvroInt().accepts(AvroLong()))
        self.assertFalse(AvroInt().accepts(AvroFloat()))
        self.assertFalse(AvroInt().accepts(AvroDouble()))

        self.assertFalse(AvroLong().accepts(AvroFloat()))
        self.assertFalse(AvroLong().accepts(AvroDouble()))

        self.assertFalse(AvroFloat().accepts(AvroDouble()))

    def testPromoteNumbersTypeVariantArray(self):
        self.assertTrue(AvroArray(AvroInt()).accepts(AvroArray(AvroInt())))
        self.assertTrue(AvroArray(AvroLong()).accepts(AvroArray(AvroInt())))
        self.assertTrue(AvroArray(AvroFloat()).accepts(AvroArray(AvroInt())))
        self.assertTrue(AvroArray(AvroDouble()).accepts(AvroArray(AvroInt())))

        self.assertTrue(AvroArray(AvroLong()).accepts(AvroArray(AvroLong())))
        self.assertTrue(AvroArray(AvroFloat()).accepts(AvroArray(AvroLong())))
        self.assertTrue(AvroArray(AvroDouble()).accepts(AvroArray(AvroLong())))

        self.assertTrue(AvroArray(AvroFloat()).accepts(AvroArray(AvroFloat())))
        self.assertTrue(AvroArray(AvroDouble()).accepts(AvroArray(AvroFloat())))

        self.assertTrue(AvroArray(AvroDouble()).accepts(AvroArray(AvroDouble())))

        self.assertFalse(AvroArray(AvroInt()).accepts(AvroArray(AvroLong())))
        self.assertFalse(AvroArray(AvroInt()).accepts(AvroArray(AvroFloat())))
        self.assertFalse(AvroArray(AvroInt()).accepts(AvroArray(AvroDouble())))

        self.assertFalse(AvroArray(AvroLong()).accepts(AvroArray(AvroFloat())))
        self.assertFalse(AvroArray(AvroLong()).accepts(AvroArray(AvroDouble())))

        self.assertFalse(AvroArray(AvroFloat()).accepts(AvroArray(AvroDouble())))

    def testPromoteNumbersTypeVariantMap(self):
        self.assertTrue(AvroMap(AvroInt()).accepts(AvroMap(AvroInt())))
        self.assertTrue(AvroMap(AvroLong()).accepts(AvroMap(AvroInt())))
        self.assertTrue(AvroMap(AvroFloat()).accepts(AvroMap(AvroInt())))
        self.assertTrue(AvroMap(AvroDouble()).accepts(AvroMap(AvroInt())))

        self.assertTrue(AvroMap(AvroLong()).accepts(AvroMap(AvroLong())))
        self.assertTrue(AvroMap(AvroFloat()).accepts(AvroMap(AvroLong())))
        self.assertTrue(AvroMap(AvroDouble()).accepts(AvroMap(AvroLong())))

        self.assertTrue(AvroMap(AvroFloat()).accepts(AvroMap(AvroFloat())))
        self.assertTrue(AvroMap(AvroDouble()).accepts(AvroMap(AvroFloat())))

        self.assertTrue(AvroMap(AvroDouble()).accepts(AvroMap(AvroDouble())))

        self.assertFalse(AvroMap(AvroInt()).accepts(AvroMap(AvroLong())))
        self.assertFalse(AvroMap(AvroInt()).accepts(AvroMap(AvroFloat())))
        self.assertFalse(AvroMap(AvroInt()).accepts(AvroMap(AvroDouble())))

        self.assertFalse(AvroMap(AvroLong()).accepts(AvroMap(AvroFloat())))
        self.assertFalse(AvroMap(AvroLong()).accepts(AvroMap(AvroDouble())))

        self.assertFalse(AvroMap(AvroFloat()).accepts(AvroMap(AvroDouble())))

    def testPromoteNumbersInRecord(self):
        self.assertTrue(AvroRecord([AvroField("one", AvroInt())], name="One").accepts(AvroRecord([AvroField("one", AvroInt())], name="One")))
        self.assertTrue(AvroRecord([AvroField("one", AvroLong())], name="One").accepts(AvroRecord([AvroField("one", AvroInt())], name="One")))
        self.assertTrue(AvroRecord([AvroField("one", AvroFloat())], name="One").accepts(AvroRecord([AvroField("one", AvroInt())], name="One")))
        self.assertTrue(AvroRecord([AvroField("one", AvroDouble())], name="One").accepts(AvroRecord([AvroField("one", AvroInt())], name="One")))

        self.assertTrue(AvroRecord([AvroField("one", AvroLong())], name="One").accepts(AvroRecord([AvroField("one", AvroLong())], name="One")))
        self.assertTrue(AvroRecord([AvroField("one", AvroFloat())], name="One").accepts(AvroRecord([AvroField("one", AvroLong())], name="One")))
        self.assertTrue(AvroRecord([AvroField("one", AvroDouble())], name="One").accepts(AvroRecord([AvroField("one", AvroLong())], name="One")))

        self.assertTrue(AvroRecord([AvroField("one", AvroFloat())], name="One").accepts(AvroRecord([AvroField("one", AvroFloat())], name="One")))
        self.assertTrue(AvroRecord([AvroField("one", AvroDouble())], name="One").accepts(AvroRecord([AvroField("one", AvroFloat())], name="One")))

        self.assertTrue(AvroRecord([AvroField("one", AvroDouble())], name="One").accepts(AvroRecord([AvroField("one", AvroDouble())], name="One")))

        self.assertFalse(AvroRecord([AvroField("one", AvroInt())], name="One").accepts(AvroRecord([AvroField("one", AvroLong())], name="One")))
        self.assertFalse(AvroRecord([AvroField("one", AvroInt())], name="One").accepts(AvroRecord([AvroField("one", AvroFloat())], name="One")))
        self.assertFalse(AvroRecord([AvroField("one", AvroInt())], name="One").accepts(AvroRecord([AvroField("one", AvroDouble())], name="One")))

        self.assertFalse(AvroRecord([AvroField("one", AvroLong())], name="One").accepts(AvroRecord([AvroField("one", AvroFloat())], name="One")))
        self.assertFalse(AvroRecord([AvroField("one", AvroLong())], name="One").accepts(AvroRecord([AvroField("one", AvroDouble())], name="One")))

        self.assertFalse(AvroRecord([AvroField("one", AvroFloat())], name="One").accepts(AvroRecord([AvroField("one", AvroDouble())], name="One")))

    def testHaveTypeSafeNull(self):
        self.assertFalse(AvroNull().accepts(AvroString()))
        self.assertFalse(AvroString().accepts(AvroNull()))

    def testDistinguishBooleanFromNumbers(self):
        self.assertFalse(AvroBoolean().accepts(AvroInt()))
        self.assertFalse(AvroBoolean().accepts(AvroLong()))
        self.assertFalse(AvroBoolean().accepts(AvroFloat()))
        self.assertFalse(AvroBoolean().accepts(AvroDouble()))

        self.assertFalse(AvroInt().accepts(AvroBoolean()))
        self.assertFalse(AvroLong().accepts(AvroBoolean()))
        self.assertFalse(AvroFloat().accepts(AvroBoolean()))
        self.assertFalse(AvroDouble().accepts(AvroBoolean()))

    def testDistinguishBytesStringEnumFixed(self):
        self.assertTrue(AvroBytes().accepts(AvroBytes()))
        self.assertFalse(AvroBytes().accepts(AvroString()))
        self.assertFalse(AvroBytes().accepts(AvroEnum(["one", "two", "three"], name="One")))
        self.assertFalse(AvroBytes().accepts(AvroFixed(5, name="One")))

        self.assertFalse(AvroString().accepts(AvroBytes()))
        self.assertTrue(AvroString().accepts(AvroString()))
        self.assertFalse(AvroString().accepts(AvroEnum(["one", "two", "three"], name="One")))
        self.assertFalse(AvroString().accepts(AvroFixed(5, name="One")))

        self.assertFalse(AvroEnum(["one", "two", "three"], name="One").accepts(AvroBytes()))
        self.assertFalse(AvroEnum(["one", "two", "three"], name="One").accepts(AvroString()))
        self.assertTrue(AvroEnum(["one", "two", "three"], name="One").accepts(AvroEnum(["one", "two", "three"], name="One")))
        self.assertFalse(AvroEnum(["one", "two", "three"], name="One").accepts(AvroFixed(5, name="One")))

        self.assertFalse(AvroFixed(5, name="One").accepts(AvroBytes()))
        self.assertFalse(AvroFixed(5, name="One").accepts(AvroString()))
        self.assertFalse(AvroFixed(5, name="One").accepts(AvroEnum(["one", "two", "three"], name="One")))
        self.assertTrue(AvroFixed(5, name="One").accepts(AvroFixed(5, name="One")))

    def testResolveUnions(self):
        self.assertTrue(AvroUnion([AvroInt(), AvroString()]).accepts(AvroInt()))
        self.assertTrue(AvroUnion([AvroInt(), AvroString()]).accepts(AvroString()))
        self.assertFalse(AvroUnion([AvroInt(), AvroString()]).accepts(AvroDouble()))

        self.assertTrue(AvroUnion([AvroInt(), AvroString(), AvroNull()]).accepts(AvroUnion([AvroInt(), AvroString()])))
        self.assertTrue(AvroUnion([AvroInt(), AvroString(), AvroNull()]).accepts(AvroUnion([AvroInt(), AvroNull()])))

    def doParserTest(self, input, expected):
        found = ForwardDeclarationParser().parse([input])[input]
        self.assertTrue(found.accepts(expected) and expected.accepts(found))

    def testReadSimpleTypes(self):
        self.doParserTest('''"null"''', AvroNull())
        self.doParserTest('''"boolean"''', AvroBoolean())
        self.doParserTest('''"int"''', AvroInt())
        self.doParserTest('''"long"''', AvroLong())
        self.doParserTest('''"float"''', AvroFloat())
        self.doParserTest('''"double"''', AvroDouble())
        self.doParserTest('''"bytes"''', AvroBytes())
        self.doParserTest('''{"type": "fixed", "name": "Test", "size": 5}''', AvroFixed(5, name="Test"))
        self.doParserTest('''"string"''', AvroString())
        self.doParserTest('''{"type": "enum", "name": "Test", "symbols": ["one", "two", "three"]}''', AvroEnum(["one", "two", "three"], name="Test"))
        self.doParserTest('''{"type": "array", "items": "int"}''', AvroArray(AvroInt()))
        self.doParserTest('''{"type": "map", "values": "int"}''', AvroMap(AvroInt()))
        self.doParserTest('''{"type": "record", "name": "Test", "fields": [{"name": "one", "type": "int"}]}''', AvroRecord([AvroField("one", AvroInt())], name="Test"))
        self.doParserTest('''["string", "null"]''', AvroUnion([AvroString(), AvroNull()]))

    def testIdentifyNestedRecordDefinitions(self):
        self.maxDiff = None

        self.doParserTest(
          '''{"type": "record", "name": "Outer", "fields": [{"name": "child", "type": {"type": "record", "name": "Inner", "fields": [{"name": "child", "type": "int"}]}}]}''',
          AvroRecord([AvroField("child", AvroRecord([AvroField("child", AvroInt())], name="Inner"))], name="Outer"))

    def testResolveDependentRecordsInAnyOrder(self):
        record1 = '''{"type": "record", "name": "Outer", "fields": [{"name": "child", "type": "Inner"}]}'''
        record2 = '''{"type": "record", "name": "Inner", "fields": [{"name": "child", "type": "int"}]}'''

        expected1 = AvroRecord([AvroField("child", AvroRecord([AvroField("child", AvroInt())], name="Inner"))], name="Outer")
        expected2 = AvroRecord([AvroField("child", AvroInt())], name="Inner")

        found = ForwardDeclarationParser().parse([record1, record2])
        self.assertTrue(found[record1].accepts(expected1) and expected1.accepts(found[record1]))
        self.assertTrue(found[record2].accepts(expected2) and expected2.accepts(found[record2]))

        found = ForwardDeclarationParser().parse([record2, record1])
        self.assertTrue(found[record1].accepts(expected1) and expected1.accepts(found[record1]))
        self.assertTrue(found[record2].accepts(expected2) and expected2.accepts(found[record2]))

    def testResolveDependentRecordsWithNamespaces(self):
        record1 = '''{"type": "record", "name": "Outer", "namespace": "com.wowie", "fields": [{"name": "child", "type": "Inner"}]}'''
        record2 = '''{"type": "record", "name": "Inner", "namespace": "com.wowie", "fields": [{"name": "child", "type": "int"}]}'''

        expected1 = AvroRecord([AvroField("child", AvroRecord([AvroField("child", AvroInt())], name="Inner", namespace="com.wowie"))], name="Outer", namespace="com.wowie")
        expected2 = AvroRecord([AvroField("child", AvroInt())], name="Inner", namespace="com.wowie")

        found = ForwardDeclarationParser().parse([record1, record2])
        self.assertTrue(found[record1].accepts(expected1) and expected1.accepts(found[record1]))
        self.assertTrue(found[record2].accepts(expected2) and expected2.accepts(found[record2]))

        found = ForwardDeclarationParser().parse([record2, record1])
        self.assertTrue(found[record1].accepts(expected1) and expected1.accepts(found[record1]))
        self.assertTrue(found[record2].accepts(expected2) and expected2.accepts(found[record2]))

    def testResolveRecursivelyDefinedRecords(self):
        record1 = '''{"type": "record", "name": "Recursive", "fields": [{"name": "child", "type": ["null", "Recursive"]}]}'''
        result = ForwardDeclarationParser().parse([record1])[record1]

        x = result.field("child").avroType
        y = AvroUnion([AvroNull(), result])
        self.assertTrue(x.accepts(y) and y.accepts(x))

    def testReCallableAndAccumulateTypes(self):
        forwardDeclarationParser = ForwardDeclarationParser()

        record1 = '''{"type": "record", "name": "Outer", "fields": [{"name": "child", "type": "Inner"}]}'''
        record2 = '''{"type": "record", "name": "Inner", "fields": [{"name": "child", "type": "int"}]}'''
        record3 = '''{"type": "record", "name": "Another", "fields": [{"name": "innerChild", "type": "Inner"}]}'''

        type1 = AvroRecord([AvroField("child", AvroRecord([AvroField("child", AvroInt())], name="Inner"))], name="Outer")
        type2 = AvroRecord([AvroField("child", AvroInt())], name="Inner")
        type3 = AvroRecord([AvroField("innerChild", AvroRecord([AvroField("child", AvroInt())], name="Inner"))], name="Another")

        result1 = forwardDeclarationParser.parse([record1, record2])
        self.assertTrue(result1[record1].accepts(type1) and type1.accepts(result1[record1]))
        self.assertTrue(result1[record2].accepts(type2) and type2.accepts(result1[record2]))

        result2 = forwardDeclarationParser.parse([record3])
        self.assertTrue(result2[record3].accepts(type3) and type3.accepts(result2[record3]))

        def equivalent(x, y):
            return x.accepts(y) and y.accepts(x)

        self.assertTrue(equivalent(forwardDeclarationParser.getAvroType("Outer"), type1))
        self.assertTrue(equivalent(forwardDeclarationParser.getAvroType("Inner"), type2))
        self.assertTrue(equivalent(forwardDeclarationParser.getAvroType("Another"), type3))
        # forwardDeclarationParser.getAvroType('''"Outer"''') should be (Some(type1))
        # forwardDeclarationParser.getAvroType('''{"type": "Outer"}''') should be (Some(type1))
        # forwardDeclarationParser.getAvroType('''{"type": "array", "items": "Outer"}''') should be (Some(AvroArray(type1)))
        # forwardDeclarationParser.getAvroType('''{"type": "array", "items": "int"}''') should be (Some(AvroArray(AvroInt())))

if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
from collections import OrderedDict

class PythonToPfaException(Exception):
    """Exception for errors encountered when converting Python code into PFA."""
    pass

defaultOptions = {"lineNumbers": True,
                  "wantArray": False}

def pfa(expr, subs=None, symbols=None, cells=None, pools=None, options=None, **kwds):
    """Convert Python code into a PFA expression (not a whole PFA document).

    :type expr: string
    :param expr: Python code in a string
    :type subs: dict from substitution strings to their values
    :param subs: Python identifiers to expand to given PFA structures
    :type symbols: list of strings
    :param symbols: Python identifiers to assume are PFA expressions in scope
    :type cells: list of strings
    :param cells: Python identifiers to assume are PFA cell names
    :type pools: list of strings
    :param pools: Python identifiers to assume are PFA pool names
    :type options: dict from option names to their values
    :param options: options for interpreting the Python and producing PFA, such as ``lineNumbers`` and ``wantArray``
    :type kwds: dict from substitution strings to their values
    :param kwds: added to ``subs`` (a more convenient way to pass them)
    :rtype: Pythonized JSON
    :return: the resulting PFA expression
    """

    if subs is None:
        subs = {}
    if len(kwds) > 0:
        subs = dict(subs)
        subs.update(kwds)

    if symbols is None:
        symbols = set()
    else:
        symbols = set(symbols)
    if cells is None:
        cells = set()
    else:
        cells = set(cells)
    if pools is None:
        pools = set()
    else:
        pools = set(pools)

    combinedOptions = dict(defaultOptions)
    if options is not None:
        combinedOptions.update(options)

    state = {"subs": subs, "symbols": symbols, "cells": cells, "pools": pools, "options": combinedOptions}
    out = _statements(ast.parse(expr).body, state)
    if not combinedOptions["wantArray"] and len(out) == 1:
        out = out[0]
    return out

def fcn(params, ret, expr, subs=None, cells=None, pools=None, options=None, **kwds):
    """Convert Python code into a PFA function (not a whole PFA document).

    :type params: Pythonized JSON
    :param params: PFA function params, form is ``[{"argName1": argType1}, {"argName2": argType2}, ...]`` where the ``argTypes`` are Avro type schemas
    :type ret: Pythonized JSON
    :param ret: PFA function ret, form is Avro type schema
    :type subs: dict from substitution strings to their values
    :param subs: Python identifiers to expand to given PFA structures
    :type cells: list of strings
    :param cells: Python identifiers to assume are PFA cell names
    :type pools: list of strings
    :param pools: Python identifiers to assume are PFA pool names
    :type options: dict from option names to their values
    :param options: options for interpreting the Python and producing PFA, such as ``lineNumbers`` and ``wantArray``
    :type kwds: dict from substitution strings to their values
    :param kwds: added to ``subs`` (a more convenient way to pass them)
    :rtype: Pythonized JSON
    :return: the resulting PFA function
    """
    combinedOptions = dict(defaultOptions)
    if options is not None:
        combinedOptions.update(options)
    return OrderedDict([("params", params), ("ret", ret), ("do", pfa(expr, subs, None, cells, pools, options, **kwds))])

def _form(state, lineno, others):
    if state["options"]["lineNumbers"] and lineno is not None:
        return OrderedDict([("@", "Python line {0}".format(lineno))] + list(others.items()))
    else:
        return others

def _newscope(state, newvars=None):
    newstate = dict(state)
    newstate["symbols"] = set(newstate["symbols"])
    if newvars is not None:
        for x in newvars:
            newstate["symbols"].add(x)
    return newstate

def _statements(stmts, state):
    return [_statement(x, state) for x in stmts]

def _statement(stmt, state):
    if isinstance(stmt, ast.Assign):
        if len(stmt.targets) == 1 and isinstance(stmt.targets[0], ast.Name):
            name = stmt.targets[0].id
            expr = _expression(stmt.value, state)
            if name in state["symbols"]:
                letset = "set"
            else:
                letset = "let"
            state["symbols"].add(name)
            return _form(state, stmt.lineno, {letset: {name: expr}})

        elif len(stmt.targets) == 1 and isinstance(stmt.targets[0], ast.Tuple) and all(isinstance(x, ast.Name) for x in stmt.targets[0].elts) and isinstance(stmt.value, ast.Tuple):
            out = OrderedDict()
            lets = 0
            sets = 0
            for name, value in zip(stmt.targets[0].elts, stmt.value.elts):
                if name.id in state["symbols"]:
                    sets += 1
                else:
                    lets += 1
                    state["symbols"].add(name.id)
                out[name.id] = _expression(value, state)
            if lets > 0 and sets == 0:
                letset = "let"
            elif lets == 0 and sets > 0:
                letset = "set"
            else:
                raise PythonToPfaException("cannot declare {0} symbols and reassign {1} in the same statement (source line {2})".format(lets, sets, stmt.lineno))
            return _form(state, stmt.lineno, {letset: out})

    elif isinstance(stmt, ast.Print):
        return _form(state, stmt.lineno, {"log": [_expression(x, state) for x in stmt.values]})

    elif isinstance(stmt, ast.For):
        if isinstance(stmt.target, ast.Name):
            newvar = stmt.target.id
            array = _expression(stmt.iter, _newscope(state))
            body = _statements(stmt.body, _newscope(state, [newvar]))
            return _form(state, stmt.lineno, OrderedDict([("foreach", newvar), ("in", array), ("do", body), ("seq", True)]))

        elif isinstance(stmt.target, ast.Tuple) and len(stmt.target.elts) == 2 and all(isinstance(x, ast.Name) for x in stmt.target.elts):
            keyvar = stmt.target.elts[0].id
            valvar = stmt.target.elts[1].id
            mapping = _expression(stmt.iter, _newscope(state))
            body = _statements(stmt.body, _newscope(state, [keyvar, valvar]))
            return _form(state, stmt.lineno, OrderedDict([("forkey", keyvar), ("forval", valvar), ("in", mapping), ("do", body)]))
            
    elif isinstance(stmt, ast.While):
        test = _expression(stmt.test, _newscope(state))
        body = _statements(stmt.body, _newscope(state))
        return _form(state, stmt.lineno, OrderedDict([("while", test), ("do", body)]))

    elif isinstance(stmt, ast.If):
        chain = _ifchain(state, stmt, [])
        if len(chain) == 1:
            return _form(state, stmt.lineno, chain[0])
        elif len(chain) == 2 and not isinstance(chain[1], OrderedDict):
            chain[0]["else"] = chain[1]
            return _form(state, stmt.lineno, chain[0])
        elif not isinstance(chain[-1], OrderedDict):
            ifThens, elseClause = chain[:-1], chain[-1]
            return _form(state, stmt.lineno, OrderedDict([("cond", ifThens), ("else", elseClause)]))
        else:
            return _form(state, stmt.lineno, OrderedDict([("cond", chain)]))

    elif isinstance(stmt, ast.Raise):
        if isinstance(stmt.type, ast.Str):
            return _form(state, stmt.lineno, {"error": stmt.type.s})

    elif isinstance(stmt, ast.Expr):
        return _expression(stmt.value, state)

    elif isinstance(stmt, ast.Pass):
        return _form(state, stmt.lineno, {"doc": ""})

    raise PythonToPfaException("Python AST node {0} (source line {1}) does not have a PFA equivalent".format(ast.dump(stmt), stmt.lineno))

def _ifchain(state, stmt, chain):
    test = _expression(stmt.test, _newscope(state))
    thenClause = _statements(stmt.body, _newscope(state))
    chain.append(OrderedDict([("if", test), ("then", thenClause)]))

    if len(stmt.orelse) == 0:
        return chain

    elif len(stmt.orelse) == 1 and isinstance(stmt.orelse[0], ast.If):
        return _ifchain(state, stmt.orelse[0], chain)

    else:
        chain.append(_statements(stmt.orelse, _newscope(state)))
        return chain

def _expression(expr, state):
    if isinstance(expr, ast.BoolOp):
        if isinstance(expr.op, ast.And):
            andor = "&&"
        else:
            andor = "||"

        items = list(reversed(expr.values))
        out = _form(state, expr.lineno, {andor: [_expression(items.pop(), _newscope(state)), _expression(items.pop(), _newscope(state))]})

        while len(items) > 0:
            out = _form(state, expr.lineno, {andor: [out, _expression(items.pop(), _newscope(state))]})

        return out

    elif isinstance(expr, ast.BinOp):
        if isinstance(expr.op, ast.Add):
            op = "+"
        elif isinstance(expr.op, ast.Sub):
            op = "-"
        elif isinstance(expr.op, ast.Mult):
            op = "*"
        elif isinstance(expr.op, ast.Div):
            op = "/"
        elif isinstance(expr.op, ast.Mod):
            op = "%"
        elif isinstance(expr.op, ast.Pow):
            op = "**"
        elif isinstance(expr.op, ast.LShift):
            raise PythonToPfaException("Python AST node {0} (source line {1}) does not have a PFA equivalent".format(ast.dump(expr), expr.lineno))
        elif isinstance(expr.op, ast.RShift):
            raise PythonToPfaException("Python AST node {0} (source line {1}) does not have a PFA equivalent".format(ast.dump(expr), expr.lineno))
        elif isinstance(expr.op, ast.BitOr):
            op = "|"
        elif isinstance(expr.op, ast.BitXor):
            op = "^"
        elif isinstance(expr.op, ast.BitAnd):
            op = "&"
        elif isinstance(expr.op, ast.FloorDiv):
            op = "//"
        return _form(state, expr.lineno, {op: [_expression(expr.left, _newscope(state)), _expression(expr.right, _newscope(state))]})

    elif isinstance(expr, ast.UnaryOp):
        if isinstance(expr.op, ast.Invert):
            op = "~"
        elif isinstance(expr.op, ast.Not):
            op = "!"
        elif isinstance(expr.op, ast.UAdd):
            raise PythonToPfaException("Python AST node {0} (source line {1}) does not have a PFA equivalent".format(ast.dump(expr), expr.lineno))
        elif isinstance(expr.op, ast.USub):
            op = "u-"
        return _form(state, expr.lineno, {op: _expression(expr.operand, _newscope(state))})

    elif isinstance(expr, ast.IfExp):
        test = _expression(expr.test, _newscope(state))
        thenClause = _expression(expr.body, _newscope(state))
        elseClause = _expression(expr.orelse, _newscope(state))
        return _form(state, expr.lineno, OrderedDict([("if", test), ("then", thenClause), ("else", elseClause)]))

    elif isinstance(expr, ast.Dict):
        return _literal(expr, state)
    
    elif isinstance(expr, ast.Compare):
        def opname(x):
            if isinstance(x, ast.Eq): return "=="
            elif isinstance(x, ast.NotEq): return "!="
            elif isinstance(x, ast.Lt): return "<"
            elif isinstance(x, ast.LtE): return "<="
            elif isinstance(x, ast.Gt): return ">"
            elif isinstance(x, ast.GtE): return ">="
            else:
                raise PythonToPfaException("Python AST node {0} (source line {1}) does not have a PFA equivalent".format(ast.dump(x), x.lineno))

        ops = list(reversed(expr.ops))
        cmps = list(reversed(expr.comparators))

        lastexpr = _expression(cmps.pop(), _newscope(state))
        out = _form(state, expr.lineno, {opname(ops.pop()): [_expression(expr.left, _newscope(state)), lastexpr]})

        while len(ops) > 0 and len(cmps) > 0:
            newexpr = _expression(cmps.pop(), _newscope(state))
            out = _form(state, expr.lineno, {"&&": [out, _form(state, expr.lineno, {opname(ops.pop()): [lastexpr, newexpr]})]})
            lastexpr = newexpr
                  
        return out

    elif isinstance(expr, ast.Call):
        def unfold(x):
            if isinstance(x, ast.Name):
                return x.id
            elif isinstance(x, ast.Attribute):
                return unfold(x.value) + "." + x.attr
            else:
                raise PythonToPfaException("Python AST node {0} (source line {1}) does not have a PFA equivalent".format(ast.dump(x), x.lineno))

        specialForms = ["Int", "Long", "Float", "Double", "String", "Base64", "Type", "Value", "New", "Do", "Let", "Set", "Attr", "Path", "To", "Cell", "Pool", "Init", "If", "Then", "Else", "Cond", "While", "Until", "For", "Step", "Foreach", "In", "Forkey", "Forval", "Cast", "Cases", "Partial", "Upcast", "As", "Ifnotnull", "Doc", "Error", "Log", "Namespace", "Params", "Ret", "Fcn"]
        coreLibrary = {"Plus": "+", "Minus": "-", "Times": "*", "Divide": "/", "FloorDivide": "//", "Negative": "U-", "Modulo": "%", "Remainder": "%%", "Pow": "**", "Comparison": "cmp", "Equal": "==", "GreaterOrEqual": ">=", "GreaterThan": ">", "NotEqual": "!=", "LessThan": "<", "LessOrEqual": "<=", "Max": "max", "Min": "min", "And": "&&", "Or": "||", "XOr": "^^", "Not": "!", "BitwiseAnd": "&", "BitwiseOr": "|", "BitwiseXOr": "^", "BitwiseNot": "~"}

        name = unfold(expr.func)
        if name in specialForms:
            name = name.lower()
        name = coreLibrary.get(name, name)

        args = [_expression(x, _newscope(state)) for x in expr.args]
        if len(args) == 1:
            args = args[0]

        if name in ["string", "base64", "value", "attr", "cell", "pool", "foreach", "forkey", "forval", "doc", "error", "fcn"] and isinstance(args, dict) and len(args) == 1 and list(args.keys()) == ["string"]:
            args, = list(args.values())

        out = OrderedDict([(name, args)])

        for kwd in expr.keywords:
            kwdarg = kwd.arg
            if kwdarg in specialForms:
                kwdarg = kwdarg.lower()
            out[kwdarg] = _expression(kwd.value, _newscope(state))

        return _form(state, expr.lineno, out)

    elif isinstance(expr, ast.Num):
        return expr.n

    elif isinstance(expr, ast.Str):
        return {"string": expr.s}

    elif isinstance(expr, ast.Attribute):
        return _unfold(expr, [], state)

    elif isinstance(expr, ast.Subscript):
        return _unfold(expr, [], state)

    elif isinstance(expr, ast.Name):
        if expr.id in state["subs"]:
            return state["subs"][expr.id]
        elif expr.id == "None":
            return None
        elif expr.id == "True":
            return True
        elif expr.id == "False":
            return False
        else:
            return expr.id

    elif isinstance(expr, ast.List):
        return _literal(expr, state)

    elif isinstance(expr, ast.Tuple):
        return [_expression(x, state) for x in expr.elts]

    raise PythonToPfaException("Python AST node {0} (source line {1}) does not have a PFA equivalent".format(ast.dump(expr), expr.lineno))

def _unfold(x, path, state):
    if isinstance(x, ast.Attribute):
        path.insert(0, {"string": x.attr})
        return _unfold(x.value, path, state)

    elif isinstance(x, ast.Subscript) and isinstance(x.slice, ast.Index):
        path.insert(0, _expression(x.slice.value, state))
        return _unfold(x.value, path, state)

    else:
        if isinstance(x, ast.Name) and x.id in state["cells"]:
            return _form(state, x.lineno, OrderedDict([("cell", x.id), ("path", path)]))
        elif isinstance(x, ast.Name) and x.id in state["pools"]:
            return _form(state, x.lineno, OrderedDict([("pool", x.id), ("path", path)]))
        else:
            return _form(state, x.lineno, OrderedDict([("attr", _expression(x, state)), ("path", path)]))

def _literal(expr, state):
    if isinstance(expr, ast.Dict):
        out = OrderedDict()
        for key, value in zip(expr.keys, expr.values):
            if isinstance(key, ast.Str) or (isinstance(key, ast.Name) and isinstance(state["subs"].get(key.id, None), str)):
                kkey = _literal(key, state)
                vvalue = _literal(value, state)
                out[kkey] = vvalue
            else:
                raise PythonToPfaException("literal JSON keys must be strings or subs identifiers, not {0} (source line {1})".format(ast.dump(expr), expr.lineno))
        return out

    elif isinstance(expr, ast.Num):
        return expr.n

    elif isinstance(expr, ast.Str):
        return expr.s

    elif isinstance(expr, ast.Name):
        if expr.id in state["subs"]:
            return state["subs"][expr.id]
        elif expr.id == "None":
            return None
        elif expr.id == "True":
            return True
        elif expr.id == "False":
            return False
        else:
            raise PythonToPfaException("identifiers ({0}) are not allowed in a literal expression, unless they are subs identifiers (source line {1})".format(ast.dump(expr), expr.lineno))

    elif isinstance(expr, ast.List):
        out = []
        for value in expr.elts:
            out.append(_literal(value, state))
        return out

    raise PythonToPfaException("Python AST node {0} (source line {1}) is not literal JSON".format(ast.dump(expr), expr.lineno))



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

from titus.fcn import Fcn
from titus.fcn import LibFcn
from titus.signature import Sig
from titus.signature import Sigs
from titus.datatype import *
from titus.errors import *
from titus.lib.core import INT_MIN_VALUE
from titus.lib.core import INT_MAX_VALUE
from titus.lib.core import LONG_MIN_VALUE
from titus.lib.core import LONG_MAX_VALUE
from titus.lib.core import FLOAT_MIN_VALUE
from titus.lib.core import FLOAT_MAX_VALUE
from titus.lib.core import DOUBLE_MIN_VALUE
from titus.lib.core import DOUBLE_MAX_VALUE
import titus.P as P

provides = {}
def provide(fcn):
    provides[fcn.name] = fcn

prefix = "parse."

#################################################################### functions

class ParseInt(LibFcn):
    name = prefix + "int"
    sig = Sig([{"str": P.String()}, {"base": P.Int()}], P.Int())
    errcodeBase = 33000
    def __call__(self, state, scope, pos, paramTypes, str, base):
        if base < 2 or base > 36:
            raise PFARuntimeException("base out of range", self.errcodeBase + 1, self.name, pos)
        try:
            out = int(str, base)
        except ValueError:
            raise PFARuntimeException("not an integer", self.errcodeBase + 0, self.name, pos)
        if out < INT_MIN_VALUE or out > INT_MAX_VALUE:
            raise PFARuntimeException("not an integer", self.errcodeBase + 0, self.name, pos)
        else:
            return out
provide(ParseInt())

class ParseLong(LibFcn):
    name = prefix + "long"
    sig = Sig([{"str": P.String()}, {"base": P.Int()}], P.Long())
    errcodeBase = 33010
    def __call__(self, state, scope, pos, paramTypes, str, base):
        if base < 2 or base > 36:
            raise PFARuntimeException("base out of range", self.errcodeBase + 1, self.name, pos)
        try:
            out = int(str, base)
        except ValueError:
            raise PFARuntimeException("not a long integer", self.errcodeBase + 0, self.name, pos)
        if out < LONG_MIN_VALUE or out > LONG_MAX_VALUE:
            raise PFARuntimeException("not a long integer", self.errcodeBase + 0, self.name, pos)
        else:
            return out
provide(ParseLong())

class ParseFloat(LibFcn):
    name = prefix + "float"
    sig = Sig([{"str": P.String()}], P.Float())
    errcodeBase = 33020
    def __call__(self, state, scope, pos, paramTypes, str):
        try:
            out = float(str)
        except ValueError:
            raise PFARuntimeException("not a single-precision float", self.errcodeBase + 0, self.name, pos)
        if math.isnan(out):
            return out
        elif math.isinf(out):
            return out
        elif out > FLOAT_MAX_VALUE:
            return float("inf")
        elif -out > FLOAT_MAX_VALUE:
            return float("-inf")
        elif abs(out) < FLOAT_MIN_VALUE:
            return 0.0
        else:
            return out
provide(ParseFloat())

class ParseDouble(LibFcn):
    name = prefix + "double"
    sig = Sig([{"str": P.String()}], P.Double())
    errcodeBase = 33030
    def __call__(self, state, scope, pos, paramTypes, str):
        try:
            out = float(str)
        except ValueError:
            raise PFARuntimeException("not a double-precision float", self.errcodeBase + 0, self.name, pos)
        if math.isnan(out):
            return out
        elif math.isinf(out):
            return out
        elif out > DOUBLE_MAX_VALUE:
            return float("inf")
        elif -out > DOUBLE_MAX_VALUE:
            return float("-inf")
        elif abs(out) < DOUBLE_MIN_VALUE:
            return 0.0
        else:
            return out
provide(ParseDouble())



#!/usr/bin/env python

# Copyright (C) 2014  Open Data ("Open Data" refers to
# one or more of the following companies: Open Data Partners LLC,
# Open Data Research LLC, or Open Data Capital LLC.)
# 
# This file is part of Hadrian.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import base64
import urllib.request, urllib.parse, urllib.error
import io
import re

from avro.datafile import DataFileReader
from avro.io import DatumReader

import titus.util
from titus.util import pos, bytesToString

from titus.pfaast import validSymbolName
from titus.pfaast import validFunctionName
from titus.pfaast import Ast
from titus.pfaast import Method
from titus.pfaast import EngineConfig
from titus.pfaast import Cell
from titus.pfaast import Pool
from titus.pfaast import Argument
from titus.pfaast import Expression
from titus.pfaast import LiteralValue
from titus.pfaast import PathIndex
from titus.pfaast import ArrayIndex
from titus.pfaast import MapIndex
from titus.pfaast import RecordIndex
from titus.pfaast import HasPath
from titus.pfaast import FcnDef
from titus.pfaast import FcnRef
from titus.pfaast import FcnRefFill
from titus.pfaast import CallUserFcn
from titus.pfaast import Call
from titus.pfaast import Ref
from titus.pfaast import LiteralNull
from titus.pfaast import LiteralBoolean
from titus.pfaast import LiteralInt
from titus.pfaast import LiteralLong
from titus.pfaast import LiteralFloat
from titus.pfaast import LiteralDouble
from titus.pfaast import LiteralString
from titus.pfaast import LiteralBase64
from titus.pfaast import Literal
from titus.pfaast import NewObject
from titus.pfaast import NewArray
from titus.pfaast import Do
from titus.pfaast import Let
from titus.pfaast import SetVar
from titus.pfaast import AttrGet
from titus.pfaast import AttrTo
from titus.pfaast import CellGet
from titus.pfaast import CellTo
from titus.pfaast import PoolGet
from titus.pfaast import PoolTo
from titus.pfaast import PoolDel
from titus.pfaast import If
from titus.pfaast import Cond
from titus.pfaast import While
from titus.pfaast import DoUntil
from titus.pfaast import For
from titus.pfaast import Foreach
from titus.pfaast import Forkeyval
from titus.pfaast import CastCase
from titus.pfaast import CastBlock
from titus.pfaast import Upcast
from titus.pfaast import IfNotNull
from titus.pfaast import BinaryFormatter
from titus.pfaast import Pack
from titus.pfaast import Unpack
from titus.pfaast import Doc
from titus.pfaast import Error
from titus.pfaast import Try
from titus.pfaast import Log
from titus.errors import PFASyntaxException
from titus.datatype import AvroTypeBuilder

def jsonToAst(jsonInput):
    """Reads PFA from serialized JSON into an abstract syntax tree.

    :type jsonInput: open JSON file, JSON string, or Pythonized JSON
    :param jsonInput: input JSON
    :rtype: titus.pfaast.EngineConfig
    :return: a PFA configuration that has passed syntax but not semantics checks
    """

    if isinstance(jsonInput, io.IOBase):
        jsonInput = jsonInput.read()
    if isinstance(jsonInput, str):
        jsonInput = json.loads(jsonInput)
    
    avroTypeBuilder = AvroTypeBuilder()
            
    result = _readEngineConfig(jsonInput, avroTypeBuilder)
    avroTypeBuilder.resolveTypes()
    return result

def yamlToAst(yamlInput):
    """Reads PFA from serialized YAML into an abstract syntax tree.

    :type yamlInput: open YAML file or YAML string
    :param yamlInput: input YAML
    :rtype: titus.pfaast.EngineConfig
    :return: a PFA configuration that has passed syntax but not semantics checks
    """
    import yaml

    def read(parser):
        try:
            while True:
                event = next(parser)
        
                if isinstance(event, yaml.ScalarEvent):
                    if not event.implicit[0]:
                        return event.value
                    elif event.value in ("yes", "Yes", "YES", "true", "True", "TRUE", "on", "On", "ON"):
                        return True
                    elif event.value in ("no", "No", "NO", "false", "False", "FALSE", "off", "Off", "OFF"):
                        return False
                    elif event.value in ("null", "Null", "NULL"):
                        return None
                    else:
                        try:
                            return int(event.value)
                        except ValueError:
                            try:
                                return float(event.value)
                            except ValueError:
                                return event.value

                elif isinstance(event, yaml.SequenceStartEvent):
                    out = []
                    while True:
                        item = read(parser)
                        if isinstance(item, yaml.SequenceEndEvent):
                            return out
                        elif isinstance(item, yaml.events.Event):
                            raise PFASyntaxException("malformed YAML", "line {0}".format(event.end_mark.line))
                        else:
                            out.append(item)

                elif isinstance(event, yaml.MappingStartEvent):
                    out = {}
                    while True:
                        startLine = event.start_mark.line
                        key = read(parser)
                        if isinstance(key, yaml.MappingEndEvent):
                            endLine = key.end_mark.line
                            if startLine == endLine:
                                out["@"] = "YAML line {0}".format(startLine)
                            else:
                                out["@"] = "YAML lines {0}-{1}".format(startLine, endLine)

                            return out

                        elif isinstance(key, yaml.events.Event):
                            raise PFASyntaxException("malformed YAML", "line {0}".format(event.end_mark.line))
                        elif not isinstance(key, str):
                            raise PFASyntaxException("YAML keys must be strings, not {0}".format(key), "line {0}".format(event.end_mark.line))

                        value = read(parser)
                        if isinstance(value, yaml.events.Event):
                            raise PFASyntaxException("malformed YAML", "line {0}".format(event.end_mark.line))
                        
                        out[key] = value

                elif isinstance(event, (yaml.SequenceEndEvent, yaml.MappingEndEvent, yaml.DocumentEndEvent, yaml.StreamEndEvent)):
                    return event

        except StopIteration:
            return event

    obj = read(yaml.parse(yamlInput, Loader=yaml.SafeLoader))
    if isinstance(obj, yaml.events.Event):
        raise PFASyntaxException("YAML document does not contain any elements that map to JSON", "")

    return jsonToAst(obj)

def jsonToExpressionAst(jsonInput, where=""):
    """Parse a PFA expression as a PFA abstract syntax tree.

    :type x: open JSON file, JSON string, or Pythonized JSON
    :param x: PFA expressions in a JSON array
    :rtype: titus.pfaast.Expression
    :return: parsed expression as a single abstract syntax tree
    """
    if isinstance(jsonInput, io.IOBase):
        jsonInput = jsonInput.read()
    if isinstance(jsonInput, str):
        jsonInput = json.loads(jsonInput)

    avroTypeBuilder = AvroTypeBuilder()

    result = _readExpression(jsonInput, where, avroTypeBuilder)
    avroTypeBuilder.resolveTypes()
    return result

jsonToAst.expr = jsonToExpressionAst

def jsonToExpressionsAst(jsonInput, where=""):
    """Parse a JSON array of PFA expressions as a PFA abstract syntax trees.

    :type x: open JSON file, JSON string, or Pythonized JSON
    :param x: PFA expressions in a JSON array
    :rtype: list of titus.pfaast.Expression
    :return: parsed expressions as a list of abstract syntax trees
    """
    if isinstance(jsonInput, io.IOBase):
        jsonInput = jsonInput.read()
    if isinstance(jsonInput, str):
        jsonInput = json.loads(jsonInput)

    avroTypeBuilder = AvroTypeBuilder()

    result = _readExpressionArray(jsonInput, where, avroTypeBuilder)
    avroTypeBuilder.resolveTypes()
    return result

jsonToAst.exprs = jsonToExpressionsAst

def jsonToFcnDef(jsonInput, where=""):
    """Parse a PFA function definition as a PFA abstract syntax tree.

    :type x: open JSON file, JSON string, or Pythonized JSON
    :param x: PFA function definition
    :rtype: titus.pfaast.FcnDef
    :return: parsed function as a single abstract syntax tree
    """
    if isinstance(jsonInput, io.IOBase):
        jsonInput = jsonInput.read()
    if isinstance(jsonInput, str):
        jsonInput = json.loads(jsonInput)

    avroTypeBuilder = AvroTypeBuilder()
    
    result = _readFcnDef(jsonInput, where, avroTypeBuilder)
    avroTypeBuilder.resolveTypes()
    return result

jsonToAst.fcn = jsonToFcnDef

def jsonToFcnDefs(jsonInput, where=""):
    """Parse a JSON object of PFA function declarations as a PFA abstract syntax tree.

    :type x: open JSON file, JSON string, or Pythonized JSON
    :param x: PFA functions in a JSON object
    :rtype: dict of titus.pfaast.FcnDef
    :return: parsed functions as abstract syntax trees
    """
    if isinstance(jsonInput, io.IOBase):
        jsonInput = jsonInput.read()
    if isinstance(jsonInput, str):
        jsonInput = json.loads(jsonInput)

    avroTypeBuilder = AvroTypeBuilder()
    
    result = _readFcnDefMap(jsonInput, where, avroTypeBuilder)
    avroTypeBuilder.resolveTypes()
    return result

jsonToAst.fcns = jsonToFcnDefs

def _trunc(x):
    if len(x) > 30:
        return x[:27] + "..."
    else:
        return x

def _stripAtSigns(data):
    if isinstance(data, dict):
        return dict((k, _stripAtSigns(v)) for k, v in list(data.items()) if k != "@")
    elif isinstance(data, (list, tuple)):
        return [_stripAtSigns(x) for x in data]
    else:
        return data

def _readEngineConfig(data, avroTypeBuilder):
    at = data.get("@")

    if not isinstance(data, dict):
        raise PFASyntaxException("PFA engine must be a JSON object, not " + _trunc(repr(data)), at)

    keys = set(x for x in list(data.keys()) if x != "@")

    _method = Method.MAP
    _begin = []
    _end = []
    _fcns = {}
    _zero = None
    _merge = None
    _cells = {}
    _pools = {}
    _randseed = None
    _doc = None
    _version = None
    _metadata = {}
    _options = {}

    for key in keys:
        if key == "name": _name = _readString(data[key], key)
        elif key == "method":
            x = _readString(data[key], key)
            if x == "map":
                _method = Method.MAP
            elif x == "emit":
                _method = Method.EMIT
            elif x == "fold":
                _method = Method.FOLD
            else:
                raise PFASyntaxException("expected one of \"map\", \"emit\", \"fold\", not \"{0}\"".format(x), at)
        elif key == "input": _input = _readAvroPlaceholder(data[key], key, avroTypeBuilder)
        elif key == "output": _output = _readAvroPlaceholder(data[key], key, avroTypeBuilder)
        elif key == "begin":
            if isinstance(data[key], (list, tuple)):
                _begin = _readExpressionArray(data[key], key, avroTypeBuilder)
            else:
                _begin = [_readExpression(data[key], key, avroTypeBuilder)]
        elif key == "action":
            if isinstance(data[key], (list, tuple)):
                _action = _readExpressionArray(data[key], key, avroTypeBuilder)
            else:
                _action = [_readExpression(data[key], key, avroTypeBuilder)]
        elif key == "end":
            if isinstance(data[key], (list, tuple)):
                _end = _readExpressionArray(data[key], key, avroTypeBuilder)
            else:
                _end = [_readExpression(data[key], key, avroTypeBuilder)]
        elif key == "fcns": _fcns = _readFcnDefMap(data[key], key, avroTypeBuilder)
        elif key == "zero": _zero = _readJsonToString(data[key], key)
        elif key == "merge":
            if isinstance(data[key], (list, tuple)):
                _merge = _readExpressionArray(data[key], key, avroTypeBuilder)
            else:
                _merge = [_readExpression(data[key], key, avroTypeBuilder)]
        elif key == "cells": _cells = _readCells(data[key], key, avroTypeBuilder)
        elif key == "pools": _pools = _readPools(data[key], key, avroTypeBuilder)
        elif key == "randseed": _randseed = _readLong(data[key], key)
        elif key == "doc": _doc = _readString(data[key], key)
        elif key == "version": _version = _readInt(data[key], key)
        elif key == "metadata": _metadata = _readStringMap(data[key], key)
        elif key == "options": _options = _readJsonNodeMap(data[key], key)
        else:
            raise PFASyntaxException("unexpected top-level field: {0}".format(key), at)

    if "name" not in keys:
        _name = titus.util.uniqueEngineName()

    required = set(["action", "input", "output"])
    if keys.intersection(required) != required:
        raise PFASyntaxException("missing top-level fields: {0}".format(", ".join(required.difference(keys))), at)
    else:
        return EngineConfig(_name, _method, _input, _output, _begin, _action, _end, _fcns, _zero, _merge, _cells, _pools, _randseed, _doc, _version, _metadata, _options, at)

def _readJsonToString(data, dot):
    return json.dumps(_stripAtSigns(data))

def _readJsonNode(data, dot):
    return _stripAtSigns(data)

def _readAvroPlaceholder(data, dot, avroTypeBuilder):
    return avroTypeBuilder.makePlaceholder(json.dumps(_stripAtSigns(data)))

def _readJsonNodeMap(data, dot):
    if isinstance(data, dict):
        return _stripAtSigns(data)
    else:
        raise PFASyntaxException("expected map of JSON objects, not " + _trunc(repr(data)), dot)

def _readJsonToStringMap(data, dot):
    if isinstance(data, dict):
        at = data.get("@")
        return dict((k, _readJsonToString(v, dot + " -> " + k)) for k, v in list(data.items()) if k != "@")
    else:
        raise PFASyntaxException("expected map of JSON objects, not " + _trunc(repr(data)), dot)

def _readBoolean(data, dot):
    if isinstance(data, bool):
        return data
    else:
        raise PFASyntaxException("expected boolean, not " + _trunc(repr(data)), dot)

def _readInt(data, dot):
    if isinstance(data, int):
        if -2147483648 <= data <= 2147483647:
            return data
        else:
            raise PFASyntaxException("int out of range: {0}".format(data), dot)
    else:
        raise PFASyntaxException("expected int, not " + _trunc(repr(data)), dot)

def _readLong(data, dot):
    if isinstance(data, int):
        if -9223372036854775808 <= data <= 9223372036854775807:
            return data
        else:
            raise PFASyntaxException("long out of range: {0}".format(data), dot)
    else:
        raise PFASyntaxException("expected long, not " + _trunc(repr(data)), dot)

def _readFloat(data, dot):
    if isinstance(data, (int, float)):
        return float(data)
    else:
        raise PFASyntaxException("expected float, not " + _trunc(repr(data)), dot)

def _readDouble(data, dot):
    if isinstance(data, (int, float)):
        return float(data)
    else:
        raise PFASyntaxException("expected double, not " + _trunc(repr(data)), dot)

def _readStringExpressionPairs(data, dot, avroTypeBuilder):
    if isinstance(data, (list, tuple)):
        out = []
        for i, item in enumerate(data):
            pair = _readExpressionMap(item, dot + " -> " + str(i), avroTypeBuilder)
            try:
                del pair["@"]
            except KeyError:
                pass
            if len(pair) != 1:
                raise PFASyntaxException("expected array of {string: expression} pairs, found map of length " + str(len(pair)), dot)
            out.extend(list(pair.items()))
        return out
    else:
        return PFASyntaxException("expected array of {string: expression} pairs, found " + _trunc(repr(data)), dot)

def _readStringPairs(data, dot):
    if isinstance(data, (list, tuple)):
        out = []
        for i, item in enumerate(data):
            try:
                pair = _readStringMap(item, dot + " -> " + str(i))
            except PFASyntaxException as err:
                raise PFASyntaxException("expected array of {string: string} pairs, found non-string map", err.pos)
            try:
                del pair["@"]
            except KeyError:
                pass
            if len(pair) != 1:
                raise PFASyntaxException("expected array of {string: string} pairs, found map of length " + str(len(pair)), dot)
            out.extend(list(pair.items()))
        return out
    else:
        raise PFASyntaxException("expected array of {string: string} pairs, found " + _trunc(repr(data)), dot)

def _readStringArray(data, dot):
    if isinstance(data, (list, tuple)):
        return [_readString(x, dot + " -> " + str(i)) for i, x in enumerate(data)]
    else:
        raise PFASyntaxException("expected array of strings, not " + _trunc(repr(data)), dot)

def _readStringOrIntArray(data, dot):
    if isinstance(data, (list, tuple)):
        return [_readStringOrInt(x, dot + " -> " + str(i)) for i, x in enumerate(data)]
    else:
        raise PFASyntaxException("expected array of strings or integers, not " + _trunc(repr(data)), dot)

def _readStringMap(data, dot):
    if isinstance(data, dict):
        at = data.get("@")
        return dict((k, _readString(v, dot + " -> " + k)) for k, v in list(data.items()) if k != "@")
    else:
        raise PFASyntaxException("expected map of strings, not " + _trunc(repr(data)), dot)

def _readString(data, dot):
    if isinstance(data, str):
        return data
    else:
        raise PFASyntaxException("expected string, not " + _trunc(repr(data)), dot)

def _readStringOrInt(data, dot):
    if isinstance(data, (str, int)):
        return data
    else:
        raise PFASyntaxException("expected string or int, not " + _trunc(repr(data)), dot)

def _readBase64(data, dot):
    if isinstance(data, str):
        return bytesToString(base64.b64decode(data))
    else:
        raise PFASyntaxException("expected base64 data, not " + _trunc(repr(data)), dot)

def _readExpressionArray(data, dot, avroTypeBuilder):
    if isinstance(data, (list, tuple)):
        return [_readExpression(x, dot + " -> " + str(i), avroTypeBuilder) for i, x in enumerate(data)]
    else:
        raise PFASyntaxException("expected array of expressions, not " + _trunc(repr(data)), dot)

def _readExpressionMap(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        return dict((k, _readExpression(v, dot + " -> " + k, avroTypeBuilder)) for k, v in list(data.items()) if k != "@")
    else:
        raise PFASyntaxException("expected map of expressions, not " + _trunc(repr(data)), dot)

def _readCastCaseArray(data, dot, avroTypeBuilder):
    if isinstance(data, (list, tuple)):
        return [_readCastCase(x, dot + " -> " + str(i), avroTypeBuilder) for i, x in enumerate(data)]
    else:
        raise PFASyntaxException("expected array of cast-cases, not " + _trunc(repr(data)), dot)

def _readCastCase(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        keys = set(x for x in list(data.keys()) if x != "@")

        for key in keys:
            if key == "as": _as = _readAvroPlaceholder(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "named": _named = _readString(data[key], dot + " -> " + key)
            elif key == "do":
                if isinstance(data[key], (list, tuple)):
                    _body = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
                else:
                    _body = [_readExpression(data[key], dot + " -> " + key, avroTypeBuilder)]
            else:
                raise PFASyntaxException("unexpected field in cast-case: {0}".format(key), pos(dot, at))

    if "named" in keys and not validSymbolName(_named):
        raise PFASyntaxException("\"{0}\" is not a valid symbol name".format(_named), pos(dot, at))

    required = set(["as", "named", "do"])
    if (keys != required):
        raise PFASyntaxException("wrong set of fields for a cast-case: \"{0}\"".format(", ".join(keys)), pos(dot, at))
    else:
        return CastCase(_as, _named, _body)

def _readExpression(data, dot, avroTypeBuilder):
    out = _readArgument(data, dot, avroTypeBuilder)
    if isinstance(out, Expression):
        return out
    else:
        raise PFASyntaxException("argument appears outside of argument list", dot)

def _readArgumentArray(data, dot, avroTypeBuilder):
    if isinstance(data, (list, tuple)):
        return [_readArgument(x, dot + " -> " + str(i), avroTypeBuilder) for i, x in enumerate(data)]
    else:
        raise PFASyntaxException("expected array of arguments, not " + _trunc(repr(data)), dot)

def _readArgumentMap(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        return dict((k, _readArgument(v, dot + " -> " + k, avroTypeBuilder)) for k, v in list(data.items()) if k != "@")
    else:
        raise PFASyntaxException("expected map of arguments, found " + _trunc(repr(data)), dot)

def _readArgument(data, dot, avroTypeBuilder):
    if data is None:
        return LiteralNull(dot)
    elif isinstance(data, bool):
        return LiteralBoolean(data, dot)
    elif isinstance(data, int):
        if -2147483648 <= data <= 2147483647:
            return LiteralInt(data, dot)
        elif -9223372036854775808 <= data <= 9223372036854775807:
            return LiteralLong(data, dot)
        else:
            raise PFASyntaxException("integer out of range: " + str(data), dot)
    elif isinstance(data, float):
        return LiteralDouble(data, dot)
    elif isinstance(data, str):
        if "." in data:
            words = data.split(".")
            ref = words[0]
            rest = words[1:]
            if not validSymbolName(ref):
                raise PFASyntaxException("\"{0}\" is not a valid symbol name".format(ref), dot)
            for i in range(len(rest)):
                try:
                    asint = int(rest[i])
                except ValueError:
                    rest[i] = LiteralString(rest[i], dot)
                else:
                    rest[i] = LiteralInt(asint, dot)
            return AttrGet(Ref(ref), rest, dot)
        elif validSymbolName(data):
            return Ref(data, dot)
        else:
            raise PFASyntaxException("\"{0}\" is not a valid symbol name".format(data), dot)

    elif isinstance(data, (list, tuple)):
        if len(data) == 1 and isinstance(data[0], str):
            return LiteralString(data[0], dot)
        else:
            raise PFASyntaxException("expecting expression, which may be [\"string\"], but no other array can be used as an expression", dot)

    elif isinstance(data, dict):
        at = data.get("@")
        keys = set(x for x in list(data.keys()) if x != "@")

        _path = []
        _seq = True
        _partial = False
        _code = 0
        _newObject = None
        _newArray = None
        _filter = None

        for key in keys:
            if key == "int": _int = _readInt(data[key], dot + " -> " + key)
            elif key == "long": _long = _readLong(data[key], dot + " -> " + key)
            elif key == "float": _float = _readFloat(data[key], dot + " -> " + key)
            elif key == "double": _double = _readDouble(data[key], dot + " -> " + key)
            elif key == "string": _string = _readString(data[key], dot + " -> " + key)
            elif key == "base64": _bytes = _readBase64(data[key], dot + " -> " + key)
            elif key == "type": _avroType = _readAvroPlaceholder(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "value": _value = _readJsonToString(data[key], dot + " -> " + key)

            elif key == "let": _let = _readExpressionMap(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "set": _set = _readExpressionMap(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "for": _forlet = _readExpressionMap(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "step": _forstep = _readExpressionMap(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "ifnotnull": _ifnotnull = _readExpressionMap(data[key], dot + " -> " + key, avroTypeBuilder)

            elif key == "do":
                if isinstance(data[key], (list, tuple)):
                    _body = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
                else:
                    _body = [_readExpression(data[key], dot + " -> " + key, avroTypeBuilder)]
            elif key == "then":
                if isinstance(data[key], (list, tuple)):
                    _thenClause = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
                else:
                    _thenClause = [_readExpression(data[key], dot + " -> " + key, avroTypeBuilder)]
            elif key == "else":
                if isinstance(data[key], (list, tuple)):
                    _elseClause = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
                else:
                    _elseClause = [_readExpression(data[key], dot + " -> " + key, avroTypeBuilder)]
            elif key == "log":
                if isinstance(data[key], (list, tuple)):
                    _log = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
                else:
                    _log = [_readExpression(data[key], dot + " -> " + key, avroTypeBuilder)]
            elif key == "path":
                _path = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "args":
                _callwithargs = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "try":
                if isinstance(data[key], (list, tuple)):
                    _trycatch = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
                else:
                    _trycatch = [_readExpression(data[key], dot + " -> " + key, avroTypeBuilder)]

            elif key == "attr": _attr = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "if": _ifPredicate = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "while": _whilePredicate = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "until": _until = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "unpack": _unpack = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "del": _dell = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)

            elif key == "cond":
                _cond = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
                if any(x.elseClause is not None for x in _cond):
                    raise PFASyntaxException("cond expression must only contain else-less if expressions", pos(dot, at))

            elif key == "cases": _cases = _readCastCaseArray(data[key], dot + " -> " + key, avroTypeBuilder)

            elif key == "foreach": _foreach = _readString(data[key], dot + " -> " + key)
            elif key == "forkey": _forkey = _readString(data[key], dot + " -> " + key)
            elif key == "forval": _forval = _readString(data[key], dot + " -> " + key)
            elif key == "fcn": _fcnref = _readString(data[key], dot + " -> " + key)
            elif key == "cell": _cell = _readString(data[key], dot + " -> " + key)
            elif key == "pool": _pool = _readString(data[key], dot + " -> " + key)

            elif key == "in": _in = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "cast": _cast = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "upcast": _upcast = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "init": _init = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "call": _callwith = _readExpression(data[key], dot + " -> " + key, avroTypeBuilder)

            elif key == "seq": _seq = _readBoolean(data[key], dot + " -> " + key)
            elif key == "partial": _partial = _readBoolean(data[key], dot + " -> " + key)

            elif key == "doc": _doc = _readString(data[key], dot + " -> " + key)
            elif key == "error": _error = _readString(data[key], dot + " -> " + key)
            elif key == "code": _code = _readInt(data[key], dot + " -> " + key)
            elif key == "namespace": _namespace = _readString(data[key], dot + " -> " + key)

            elif key == "new":
                if isinstance(data[key], dict):
                    _newObject = _readExpressionMap(data[key], dot + " -> " + key, avroTypeBuilder)
                elif isinstance(data[key], (list, tuple)):
                    _newArray = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
                else:
                    raise PFASyntaxException("\"new\" must be an object (map, record) or an array", pos(dot, at))

            elif key == "params": _params = _readParams(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "ret": _ret = _readAvroPlaceholder(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "as": _as = _readAvroPlaceholder(data[key], dot + " -> " + key, avroTypeBuilder)

            elif key == "to": _to = _readArgument(data[key], dot + " -> " + key, avroTypeBuilder)

            elif key == "fill": _fill = _readArgumentMap(data[key], dot + " -> " + key, avroTypeBuilder)

            elif key == "filter": _filter = _readStringOrIntArray(data[key], dot + " -> " + key)

            elif key == "format": _format = _readStringPairs(data[key], dot + " -> " + key)
            elif key == "pack": _pack = _readStringExpressionPairs(data[key], dot + " -> " + key, avroTypeBuilder)

            else:
                _callName = key
                if isinstance(data[key], (list, tuple)):
                    _callArgs = _readArgumentArray(data[key], dot + " -> " + key, avroTypeBuilder)
                else:
                    _callArgs = [_readArgument(data[key], dot + " -> " + key, avroTypeBuilder)]

        if "foreach" in keys and not validSymbolName(_foreach):
            raise PFASyntaxException("\"{0}\" is not a valid symbol name".format(data[keys]), pos(dot, at))
        if "forkey" in keys and not validSymbolName(_forkey):
            raise PFASyntaxException("\"{0}\" is not a valid symbol name".format(data[keys]), pos(dot, at))
        if "forval" in keys and not validSymbolName(_forval):
            raise PFASyntaxException("\"{0}\" is not a valid symbol name".format(data[keys]), pos(dot, at))
        if "fcn" in keys and not validFunctionName(_fcnref):
            raise PFASyntaxException("\"{0}\" is not a valid function name".format(data[keys]), pos(dot, at))

        if keys == set(["int"]):                             return LiteralInt(_int, pos(dot, at))
        elif keys == set(["long"]):                          return LiteralLong(_long, pos(dot, at))
        elif keys == set(["float"]):                         return LiteralFloat(_float, pos(dot, at))
        elif keys == set(["double"]):                        return LiteralDouble(_double, pos(dot, at))
        elif keys == set(["string"]):                        return LiteralString(_string, pos(dot, at))
        elif keys == set(["base64"]):                        return LiteralBase64(_bytes, pos(dot, at))
        elif keys == set(["type", "value"]):                 return Literal(_avroType, _value, pos(dot, at))

        elif keys == set(["new", "type"]) and _newObject is not None:
                                                             return NewObject(_newObject, _avroType, pos(dot, at))
        elif keys == set(["new", "type"]) and _newArray is not None:
                                                             return NewArray(_newArray, _avroType, pos(dot, at))

        elif keys == set(["do"]):                            return Do(_body, pos(dot, at))
        elif keys == set(["let"]):                           return Let(_let, pos(dot, at))
        elif keys == set(["set"]):                           return SetVar(_set, pos(dot, at))

        elif keys == set(["attr", "path"]):                  return AttrGet(_attr, _path, pos(dot, at))
        elif keys == set(["attr", "path", "to"]):            return AttrTo(_attr, _path, _to, pos(dot, at))
        elif keys == set(["cell"]) or \
             keys == set(["cell", "path"]):                  return CellGet(_cell, _path, pos(dot, at))
        elif keys == set(["cell", "to"]) or \
             keys == set(["cell", "path", "to"]):            return CellTo(_cell, _path, _to, pos(dot, at))
        elif keys == set(["pool", "path"]):                  return PoolGet(_pool, _path, pos(dot, at))
        elif keys == set(["pool", "path", "to", "init"]):    return PoolTo(_pool, _path, _to, _init, pos(dot, at))
        elif keys == set(["pool", "del"]):                   return PoolDel(_pool, _dell, pos(dot, at))

        elif keys == set(["if", "then"]):                    return If(_ifPredicate, _thenClause, None, pos(dot, at))
        elif keys == set(["if", "then", "else"]):            return If(_ifPredicate, _thenClause, _elseClause, pos(dot, at))
        elif keys == set(["cond"]):                          return Cond(_cond, None, pos(dot, at))
        elif keys == set(["cond", "else"]):                  return Cond(_cond, _elseClause, pos(dot, at))

        elif keys == set(["while", "do"]):                   return While(_whilePredicate, _body, pos(dot, at))
        elif keys == set(["do", "until"]):                   return DoUntil(_body, _until, pos(dot, at))
        elif keys == set(["for", "while", "step", "do"]):    return For(_forlet, _whilePredicate, _forstep, _body, pos(dot, at))

        elif keys == set(["foreach", "in", "do"]) or \
             keys == set(["foreach", "in", "do", "seq"]):    return Foreach(_foreach, _in, _body, _seq, pos(dot, at))
        elif keys == set(["forkey", "forval", "in", "do"]):  return Forkeyval(_forkey, _forval, _in, _body, pos(dot, at))

        elif keys == set(["cast", "cases"]) or \
             keys == set(["cast", "cases", "partial"]):      return CastBlock(_cast, _cases, _partial, pos(dot, at))
        elif keys == set(["upcast", "as"]):                  return Upcast(_upcast, _as, pos(dot, at))

        elif keys == set(["ifnotnull", "then"]):             return IfNotNull(_ifnotnull, _thenClause, None, pos(dot, at))
        elif keys == set(["ifnotnull", "then", "else"]):     return IfNotNull(_ifnotnull, _thenClause, _elseClause, pos(dot, at))

        elif keys == set(["pack"]):                          return Pack(_pack, pos(dot, at))
        elif keys == set(["unpack", "format", "then"]):      return Unpack(_unpack, _format, _thenClause, None, pos(dot, at))
        elif keys == set(["unpack", "format", "then", "else"]):
                                                             return Unpack(_unpack, _format, _thenClause, _elseClause, pos(dot, at))

        elif keys == set(["doc"]):                           return Doc(_doc, pos(dot, at))

        elif keys == set(["error"]):                         return Error(_error, None, pos(dot, at))
        elif keys == set(["error", "code"]):                 return Error(_error, _code, pos(dot, at))
        elif keys == set(["try"]) or \
             keys == set(["try", "filter"]):                 return Try(_trycatch, _filter, pos(dot, at))
        elif keys == set(["log"]):                           return Log(_log, None, pos(dot, at))
        elif keys == set(["log", "namespace"]):              return Log(_log, _namespace, pos(dot, at))

        elif keys == set(["params", "ret", "do"]):           return FcnDef(_params, _ret, _body, pos(dot, at))
        elif keys == set(["fcn"]):                           return FcnRef(_fcnref, pos(dot, at))
        elif keys == set(["fcn", "fill"]):                   return FcnRefFill(_fcnref, _fill, pos(dot, at))
        elif keys == set(["call", "args"]):                  return CallUserFcn(_callwith, _callwithargs, pos(dot, at))

        elif len(keys) == 1 and list(keys)[0] not in \
             set(["args", "as", "attr", "base64", "call", "cases", "cast", "cell", "code", "cond", "do", "doc", "double", "else",
                  "error", "fcn", "fill", "filter", "float", "for", "foreach", "forkey", "format", "forval", "if", "ifnotnull", "in", "init",
                  "int", "let", "log", "long", "namespace", "new", "pack", "params", "partial", "path", "pool", "ret", "seq",
                  "set", "step", "string", "then", "to", "try", "type", "unpack", "until", "upcast", "value", "while"]):
                                                             return Call(_callName, _callArgs, pos(dot, at))

        else: raise PFASyntaxException("unrecognized special form: {0} (not enough arguments? too many?)".format(", ".join(keys)), pos(dot, at))

    else:
        raise PFASyntaxException("expected expression, not " + _trunc(repr(data)), dot)

def _readFcnDefMap(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        for k in list(data.keys()):
            if k != "@" and not validFunctionName(k):
                raise PFASyntaxException("\"{0}\" is not a valid function name".format(k), pos(dot, at))
        return dict((k, _readFcnDef(v, dot + " -> " + k, avroTypeBuilder)) for k, v in list(data.items()) if k != "@")
    else:
        raise PFASyntaxException("expected map of function definitions, not " + _trunc(repr(data)), pos(dot, at))

def _readFcnDef(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        keys = set(x for x in list(data.keys()) if x != "@")

        for key in keys:
            if key == "params": _params = _readParams(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "ret": _ret = _readAvroPlaceholder(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "do":
                if isinstance(data[key], (list, tuple)):
                    _body = _readExpressionArray(data[key], dot + " -> " + key, avroTypeBuilder)
                else:
                    _body = [_readExpression(data[key], dot + " -> " + key, avroTypeBuilder)]
            else:
                raise PFASyntaxException("unexpected field in function definition: " + key, pos(dot, at))

        required = set(["params", "ret", "do"])
        if (keys != required):
            raise PFASyntaxException("wrong set of fields for a function definition: " + ", ".join(keys), pos(dot, at))
        else:
            return FcnDef(_params, _ret, _body, pos(dot, at))
    else:
        raise PFASyntaxException("expected function definition, not " + _trunc(repr(data)), pos(dot, at))

def _readParams(data, dot, avroTypeBuilder):
    if isinstance(data, (list, tuple)):
        return [_readParam(x, dot + " -> " + str(i), avroTypeBuilder) for i, x in enumerate(data)]
    else:
        raise PFASyntaxException("expected array of function parameters, not " + _trunc(repr(data)), dot)

def _readParam(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        keys = set(x for x in list(data.keys()) if x != "@")
        if len(keys) != 1:
            raise PFASyntaxException("function parameter name-type map should have only one pair", pos(dot, at))
        n = list(keys)[0]
        if not validSymbolName(n):
            raise PFASyntaxException("\"{0}\" is not a valid symbol name".format(n))

        t = _readAvroPlaceholder(data[n], dot + " -> " + n, avroTypeBuilder)
        return {n: t}
    else:
        raise PFASyntaxException("expected function parameter name-type singleton map, not " + _trunc(repr(data)), pos(dot, at))

def _readCells(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        for k in list(data.keys()):
            if k != "@" and not validSymbolName(k):
                raise PFASyntaxException("\"{0}\" is not a valid symbol name".format(k), pos(dot, at))
        return dict((k, _readCell(data[k], dot, avroTypeBuilder)) for k, v in list(data.items()) if k != "@")
    else:
        raise PFASyntaxException("expected map of cells, not " + _trunc(repr(data)), pos(dot, at))

def _readCell(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        _shared = False
        _rollback = False
        _source = "embedded"
        keys = set(x for x in list(data.keys()) if x != "@")
        for key in keys:
            if key == "type": _avroType = _readAvroPlaceholder(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "init": _init = _readJsonToString(data[key], dot + " -> " + key)
            elif key == "shared": _shared = _readBoolean(data[key], dot + " -> " + key)
            elif key == "rollback": _rollback = _readBoolean(data[key], dot + " -> " + key)
            elif key == "source": _source = _readString(data[key], dot + " -> " + key)
            else:
                raise PFASyntaxException("unexpected cell property: \"{0}\"".format(key), pos(dot, at))

        if ("type" not in keys) or ("init" not in keys) or (not keys.issubset(set(["type", "init", "shared", "rollback", "source"]))):
            raise PFASyntaxException("wrong set of fields for a cell: " + ", ".join(keys), pos(dot, at))
        else:
            if _source == "avro":
                url = json.loads(_init)
                if not isinstance(url, str):
                    raise PFASyntaxException("source: avro requires init to be a string", pos(dot, at))
                def getit(avroType):
                    reader = DataFileReader(urllib.request.urlopen(url), DatumReader())
                    return reader.read()
                _init = getit

            elif _source == "json":
                url = json.loads(_init)
                if not isinstance(url, str):
                    raise PFASyntaxException("source: json requires init to be a string", pos(dot, at))
                def getit(avroType):
                    if re.match("^[a-zA-Z][a-zA-Z0-9\+\-\.]*://", url) is not None:
                        return urllib.request.urlopen(url).read()
                    else:
                        return open(url).read()
                _init = getit
                
            elif _source == "embedded":
                pass
                # # which is the equivalent of...
                # value = json.loads(_init)
                # def getit(avroType):
                #     return titus.datatype.jsonDecoder(avroType, value)
                # _init = getit

            else:
                raise PFASyntaxException("unrecognized source type \"{0}\"".format(_source), pos(dot, at))

            return Cell(_avroType, _init, _shared, _rollback, _source, pos(dot, at))
    else:
        raise PFASyntaxException("expected cell, not " + _trunc(repr(data)), None)

def _readPools(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        for k in list(data.keys()):
            if k != "@" and not validSymbolName(k):
                raise PFASyntaxException("\"{0}\" is not a valid symbol name".format(k), pos(dot, at))
        return dict((k, _readPool(data[k], dot, avroTypeBuilder)) for k, v in list(data.items()) if k != "@")
    else:
        raise PFASyntaxException("expected map of pools, not " + _trunc(repr(data)), None)

def _readPool(data, dot, avroTypeBuilder):
    if isinstance(data, dict):
        at = data.get("@")
        _init = {}
        _shared = False
        _rollback = False
        _source = "embedded"
        keys = set(x for x in list(data.keys()) if x != "@")
        for key in keys:
            if key == "type": _avroType = _readAvroPlaceholder(data[key], dot + " -> " + key, avroTypeBuilder)
            elif key == "init": _init = _readJsonToStringMap(data[key], dot + " -> " + key)
            elif key == "shared": _shared = _readBoolean(data[key], dot + " -> " + key)
            elif key == "rollback": _rollback = _readBoolean(data[key], dot + " -> " + key)
            elif key == "source": _source = _readString(data[key], dot + " -> " + key)

        if ("type" not in keys) or (not keys.issubset(set(["type", "init", "shared", "rollback", "source"]))):
            raise PFASyntaxException("wrong set of fields for a pool: " + ", ".join(keys), pos(dot, at))
        else:
            if _source == "avro":
                url = json.loads(_init)
                if not isinstance(url, str):
                    raise PFASyntaxException("source: avro requires init to be a string", pos(dot, at))
                def getit(avroType):
                    reader = DataFileReader(urllib.request.urlopen(url), DatumReader())
                    return reader.read()
                _init = getit

            elif _source == "json":
                url = json.loads(_init)
                if not isinstance(url, str):
                    raise PFASyntaxException("source: json requires init to be a string", pos(dot, at))
                def getit(avroType):
                    if re.match("^[a-zA-Z][a-zA-Z0-9\+\-\.]*://", url) is not None:
                        return urllib.request.urlopen(url).read()
                    else:
                        return open(url).read()
                _init = getit
                
            elif _source == "embedded":
                pass
                # # which is the equivalent of...
                # value = {}
                # for k, v in _init.items():
                #     value[k] = json.loads(v)
                # def getit(avroType):
                #     return titus.datatype.jsonDecoder(avroType, value)
                # _init = getit

            else:
                raise PFASyntaxException("unrecognized source type \"{0}\"".format(_source), pos(dot, at))

            return Pool(_avroType, _init, _shared, _rollback, _source, pos(dot, at))
    else:
        raise PFASyntaxException("expected pool, not " + _trunc(repr(data)), pos(dot, at))
